module.exports = Torrent

/* global indexedDB */

const parseTorrent = require('parse-torrent-file')
const path = require('path')
const IdbKvStore = require('idb-kv-store')
const IdbChunkStore = require('indexeddb-chunk-store')
const File = require('./file')

function Torrent (obj, namespace) {
  this.closed = false
  this.torrentMetaBuffer = Buffer.isBuffer(obj.torrentMetaBuffer)
                           ? obj.torrentMetaBuffer : new Buffer(obj.torrentMetaBuffer)
  let torrentMeta = parseTorrent(this.torrentMetaBuffer)

  this.name = torrentMeta.name
  this.created = torrentMeta.created
  this.createdBy = torrentMeta.createdBy
  this.announce = torrentMeta.announce
  this.urlList = obj.webseeds.concat(torrentMeta.urlList)
  this.length = torrentMeta.length
  this.pieceLength = torrentMeta.pieceLength
  this.lastPieceLength = torrentMeta.lastPieceLength
  this.pieces = torrentMeta.pieces
  this.infoHash = obj.infoHash
  this.files = torrentMeta.files.map(f => new File(this, f))

  this._priorityDbName = namespace + '-priority-' + this.infoHash
  this._priority = new IdbKvStore(this._priorityDbName) // TODO lazily initiaize resources
  this._missingChunks = {}
  this._chunkStore = new IdbChunkStore(this.pieceLength, {
    name: this.infoHash
  })
  this._chunkStore._store.on('set', this._onChunkPut.bind(this))
}

Torrent.prototype.getFile = function (fpath) {
  if (this.closed) throw new Error('Torrent is closed')
  fpath = path.normalize(fpath)
  if (fpath.startsWith('/')) fpath = fpath.substr(1)
  return this.files.find(f => f.path === fpath) || undefined
}

Torrent.prototype.close = function () {
  if (this.closed) return
  this.closed = true

  this.files = null

  this._priority.close()
  this._priority = null

  this._chunkStore.close()
  this._chunkStore = null
  this._missingChunks = null
}

Torrent.prototype.destroy = function () {
  this.close()
  indexedDB.deleteDatabase(this.infoHash)
  indexedDB.deleteDatabase(this._priorityDbName)
}

Torrent.prototype._onChunkPut = function (change) {
  if (this._closed) return
  if (this._missingChunks[change.key] != null) {
    let cbList = this._missingChunks[change.key]
    delete this._missingChunks[change.key]
    cbList.forEach(cb => cb(null))
  }
}

Torrent.prototype._notifyOnChunkPut = function (index, cb) {
  if (this._closed) return
  this._missingChunks[index] = this._missingChunks[index] || []
  this._missingChunks[index].push(cb)
}
